@@ -1,8 +1,6 @@
-from .classes import *
-from .path import *
 from .util import *
 from .colgen import *
 from .dtaapi import *
 
 # print out the current version
-print('path4gmns, version 0.5.2')
\ No newline at end of file
+print('path4gmns, version 0.6.0')
\ No newline at end of file
