@@ -1,58 +1,78 @@
 import ctypes
+from time import time
+
 from .path import single_source_shortest_path
-from .classes import Column, ColumnVec, MAX_TIME_PERIODS, MAX_AGNET_TYPES, \
-                     MIN_OD_VOL
+from .classes import Column, ColumnVec
 
 
 __all__ = ['perform_network_assignment']
 
 
-def _update_generalized_link_cost(links, link_cost_array, 
-                                  tau=0, value_of_time=10):
-    for link in links:
-        link_cost_array[link.get_seq_no()] = (
+_MIN_OD_VOL = 0.000001
+
+
+def _update_generalized_link_cost(spnetworks):
+    """ update generalized link costs for each SPNetwork
+
+    warning: there would be duplicate updates for SPNetworks with the same tau and vot
+    belonging to different memory blocks. It could be resolved by creating
+    a shared network among them??
+    """
+    for sp in spnetworks:
+        tau = sp.get_demand_period().get_id()
+        vot = sp.get_agent_type().get_vot()
+
+        for link in sp.get_links():
+            sp.link_cost_array[link.get_seq_no()] = (
             link.get_period_travel_time(tau)
-            + link.get_route_choice_cost() 
-            + link.get_toll() / value_of_time * 60
+            + link.get_route_choice_cost()
+            + link.get_toll() / vot * 60
         )
 
 
-def _update_link_travel_time_and_cost(links):
+def _update_link_travel_time_and_cost(links, agent_types, demand_periods):
     for link in links:
         link.calculate_td_vdfunction()
-        for tau in range(MAX_TIME_PERIODS):
-            for at in range(MAX_AGNET_TYPES):
+        for dp in demand_periods:
+            tau = dp.get_id()
+            for at in agent_types:
                 link.calculate_agent_marginal_cost(tau, at)
 
 
-def _reset_and_update_link_vol_based_on_columns(column_pool, 
+def _reset_and_update_link_vol_based_on_columns(column_pool,
                                                 links,
                                                 zones,
+                                                agent_types,
+                                                demand_periods,
                                                 iter_num,
                                                 is_path_vol_self_reducing):
+
     # the original implementation is iter_num < 0, which does not make sense
-    if iter_num <= 0:
+    if iter_num == 0:
         return
-    
+
     for link in links:
-        for tau in range(MAX_TIME_PERIODS):
+        for dperiod in demand_periods:
+            tau = dperiod.get_id()
             link.reset_period_flow_vol(tau)
             # link.queue_length_by_slot[tau] = 0
-            for at in range(MAX_AGNET_TYPES):
-                link.reset_period_agent_vol(tau, at)
+            for atype in agent_types:
+                link.reset_period_agent_vol(tau, atype.get_id())
 
-    for at in range(MAX_AGNET_TYPES):
-        for orig_zone_id in zones:
-            for dest_zone_id in zones:
-                for tau in range(MAX_TIME_PERIODS):
-                    if (orig_zone_id, dest_zone_id) not in column_pool.keys():
+    for oz_id in zones:
+        for dz_id in zones:
+            for atype in agent_types:
+                at = atype.get_id()
+                for dperiod in demand_periods:
+                    tau = dperiod.get_id()
+                    if (at, tau, oz_id, dz_id) not in column_pool.keys():
                         continue
 
-                    cv = column_pool[(orig_zone_id, dest_zone_id)]
-            
+                    cv = column_pool[(at, tau, oz_id, dz_id)]
+
                     if cv.get_od_volume() <= 0:
                         continue
-                    
+
                     for col in cv.get_columns().values():
                         link_vol_contributed_by_path_vol = col.get_volume()
                         for i in col.links:
@@ -71,32 +91,44 @@ def _reset_and_update_link_vol_based_on_columns(column_pool,
                            and is_path_vol_self_reducing:
                             col.vol *= iter_num / (iter_num + 1)
                     # end of for column in ...
-                # end of for tau in ...
-            # end of for dest_zone_id in ...
-        # end of for origi_zone_id in ...
-    # end of for at in range ...
+                # end of for dperiod in ...
+            # end of for atype in ...
+        # end of for dz_id in ...
+    # end of for oz_id in range ...
 
 
-def _update_column_gradient_cost_and_flow(column_pool, links, zones, iter_num):
-    # total_gap_count = 0
-    
-    _reset_and_update_link_vol_based_on_columns(column_pool, 
+def _update_column_gradient_cost_and_flow(column_pool,
+                                          links,
+                                          zones,
+                                          agent_types,
+                                          demand_periods,
+                                          iter_num):
+
+    _reset_and_update_link_vol_based_on_columns(column_pool,
                                                 links,
                                                 zones,
+                                                agent_types,
+                                                demand_periods,
                                                 iter_num,
                                                 False)
 
-    _update_link_travel_time_and_cost(links)
+    _update_link_travel_time_and_cost(links, agent_types, demand_periods)
 
-    for orig_zone_id in zones:
-        for dest_zone_id in zones:
-            for at in range(MAX_AGNET_TYPES):
-                for tau in range(MAX_TIME_PERIODS):
-                    if (orig_zone_id, dest_zone_id) not in column_pool.keys():
+    total_gap = 0
+    # total_gap_count = 0
+
+    for oz_id in zones:
+        for dz_id in zones:
+            for atype in agent_types:
+                at = atype.get_id()
+                vot = atype.get_vot()
+                for dperiod in demand_periods:
+                    tau = dperiod.get_id()
+                    if (at, tau, oz_id, dz_id) not in column_pool.keys():
                         continue
-                    
-                    cv = column_pool[(orig_zone_id, dest_zone_id)]
-                    
+
+                    cv = column_pool[(at, tau, oz_id, dz_id)]
+
                     if cv.get_od_volume() <= 0:
                         continue
 
@@ -109,13 +141,14 @@ def _update_column_gradient_cost_and_flow(column_pool, links, zones, iter_num):
                         path_toll = 0
                         path_gradient_cost = 0
                         path_travel_time = 0
+                        # i is link sequence no
                         for i in col.get_links():
                             path_toll += links[i].get_toll()
                             path_travel_time += (
                                 links[i].travel_time_by_period[tau]
                             )
                             path_gradient_cost += (
-                                links[i].get_generalized_cost(tau, at)
+                                links[i].get_generalized_cost(tau, vot)
                             )
 
                         col.set_toll(path_toll)
@@ -139,23 +172,32 @@ def _update_column_gradient_cost_and_flow(column_pool, links, zones, iter_num):
                         for node_sum, col in cv.get_columns().items():
                             if col.get_seq_no() != least_gradient_cost_path_seq_no:
                                 col.set_gradient_cost_abs_diff(
-                                    col.get_gradient_cost() 
+                                    col.get_gradient_cost()
                                     - least_gradient_cost
                                 )
                                 col.set_gradient_cost_rel_diff(
-                                    col.get_gradient_cost_abs_diff() 
+                                    col.get_gradient_cost_abs_diff()
                                     / max(0.0001, least_gradient_cost)
                                 )
 
+                                total_gap += (
+                                    col.get_gradient_cost_abs_diff()
+                                    * col.get_volume()
+                                )
+
+                                # total_gap_count += (
+                                #     col.get_gradient_cost() * col.get_volume()
+                                # )
+
                                 step_size = (
                                     1 / (iter_num + 2) * cv.get_od_volume()
                                 )
 
                                 previous_path_vol = col.get_volume()
                                 col.vol = max(
-                                    0, 
-                                    (previous_path_vol 
-                                     - step_size 
+                                    0,
+                                    (previous_path_vol
+                                     - step_size
                                      * col.get_gradient_cost_rel_diff())
                                 )
 
@@ -170,13 +212,27 @@ def _update_column_gradient_cost_and_flow(column_pool, links, zones, iter_num):
                         col = cv.get_column(
                             least_gradient_cost_path_node_sum
                         )
-                        col.increase_volume(total_switched_out_path_vol)                 
+                        col.increase_volume(total_switched_out_path_vol)
+
+    print(f'total gap: {total_gap:.2f}')
+    # print(f'total gap count is: {total_gap_count:.2f}')
 
 
-def _optimize_column_pool(column_pool, links, zones, colum_update_num):
+def _optimize_column_pool(column_pool,
+                          links,
+                          zones,
+                          agent_types,
+                          demand_periods,
+                          colum_update_num):
+
     for i in range(colum_update_num):
         print(f"current iteration number in column generation: {i}")
-        _update_column_gradient_cost_and_flow(column_pool, links, zones, i)
+        _update_column_gradient_cost_and_flow(column_pool,
+                                              links,
+                                              zones,
+                                              agent_types,
+                                              demand_periods,
+                                              i)
 
 
 def _backtrace_shortest_path_tree(orig_node_no,
@@ -185,13 +241,15 @@ def _backtrace_shortest_path_tree(orig_node_no,
                                   node_preds,
                                   link_preds,
                                   node_label_costs,
+                                  agent_type,
+                                  demand_period,
                                   column_pool,
                                   iter_num):
-    
+
     if not nodes[orig_node_no].has_outgoing_links():
         return
 
-    orig_zone_id = nodes[orig_node_no].get_zone_id()
+    oz_id = nodes[orig_node_no].get_zone_id()
     k_path_prob = 1 / (iter_num + 1)
 
     node_size = len(nodes)
@@ -200,21 +258,21 @@ def _backtrace_shortest_path_tree(orig_node_no,
         if i == orig_node_no:
             continue
 
-        dest_zone_id = nodes[i].get_zone_id()
-        if dest_zone_id == -1:
-            continue
-        
-        if (orig_zone_id, dest_zone_id) not in column_pool.keys():
+        dz_id = nodes[i].get_zone_id()
+        if dz_id == -1:
             continue
 
-        cv = column_pool[(orig_zone_id, dest_zone_id)]
+        if (agent_type, demand_period, oz_id, dz_id) not in column_pool.keys():
+            continue
+
+        cv = column_pool[(agent_type, demand_period, oz_id, dz_id)]
         if cv.is_route_fixed():
             continue
 
         od_vol = cv.get_od_volume()
-        if od_vol <= MIN_OD_VOL:
+        if od_vol <= _MIN_OD_VOL:
             continue
-        
+
         vol = od_vol * k_path_prob
 
         node_path = []
@@ -227,14 +285,14 @@ def _backtrace_shortest_path_tree(orig_node_no,
         while current_node_seq_no >= 0:
             node_path.append(current_node_seq_no)
             node_sum += current_node_seq_no
-            
-            current_link_seq_no = link_preds[current_node_seq_no]  
+
+            current_link_seq_no = link_preds[current_node_seq_no]
             if current_link_seq_no >= 0:
                 link_path.append(current_link_seq_no)
                 dist += links[current_link_seq_no].length
-            
+
             current_node_seq_no = node_preds[current_node_seq_no]
-        
+
         # make sure this is a valid path
         if not link_path:
             continue
@@ -251,37 +309,68 @@ def _backtrace_shortest_path_tree(orig_node_no,
         cv.get_column(node_sum).increase_volume(vol)
 
 
-def _update_column_travel_time(links, zones, column_pool):
-    for orig_zone in zones:
-        for dest_zone in zones:
-            for at in range(MAX_AGNET_TYPES):
-                for tau in range(MAX_TIME_PERIODS):
-                    if (orig_zone, dest_zone) not in column_pool.keys():
+def _update_column_travel_time(column_pool, 
+                               links,
+                               zones,
+                               agent_types,
+                               demand_periods):
+
+    for oz in zones:
+        for dz in zones:
+            for atype in agent_types:
+                at = atype.get_id()
+                for dperiod in demand_periods:
+                    dp = dperiod.get_id()
+                    if (at, dp, oz, dz) not in column_pool.keys():
                         continue
-                    
-                    cv = column_pool[(orig_zone, dest_zone)]
+
+                    cv = column_pool[(at, dp, oz, dz)]
 
                     for col in cv.get_columns().values():
                         travel_time = sum(
-                            links[j].travel_time_by_period[tau] for j in col.links
+                            links[j].travel_time_by_period[dp] for j in col.links
                         )
                         col.set_travel_time(travel_time)
-            
 
-def perform_network_assignment(assignment_mode, iter_num, column_update_num, G):
+
+def _assginment_core(spn, column_pool, iter_num):
+
+    for node_id in spn.get_orig_nodes():
+        single_source_shortest_path(spn, node_id)
+
+        _backtrace_shortest_path_tree(spn.get_node_no(node_id),
+                                      spn.get_nodes(),
+                                      spn.get_links(),
+                                      spn.get_node_preds(),
+                                      spn.get_link_preds(),
+                                      spn.get_node_label_costs(),
+                                      spn.get_agent_type().get_id(),
+                                      spn.get_demand_period().get_id(),
+                                      column_pool,
+                                      iter_num)
+
+
+def _assignment(spnetworks, column_pool, iter_num):
+    # single processing
+    # it could be multiprocessing
+    for spn in spnetworks:
+        _assginment_core(spn, column_pool, iter_num)
+
+
+def perform_network_assignment(assignment_mode, iter_num, column_update_num, ui):
     """ perform network assignemnt using the selected assignment mode
-    
+
     WARNING
     -------
         Only Path/Column-based User Equilibrium (UE) is implemented in Python.
         If you need other assignment modes or dynamic traffic assignment (DTA),
         please use perform_network_assignment_DTALite()
-    
+
     Parameters
     ----------
     assignment_mode
         0: Link-based UE
-        1: Path-based UE 
+        1: Path-based UE
         2: UE + dynamic traffic assignment and simulation
         3: ODME
     iter_num
@@ -295,53 +384,58 @@ def perform_network_assignment(assignment_mode, iter_num, column_update_num, G):
         None
 
         You will need to call output_columns() and output_link_performance() to
-        get the assignment results, i.e., paths/columns (in agent.csv) and 
+        get the assignment results, i.e., paths/columns (in agent.csv) and
         assigned volumes and other link attributes on each link (in l
         ink_performance.csv)
     """
     if assignment_mode != 1:
         raise Exception("not implemented yet")
 
-    # this is important. 
-    # otherwise, at iteration 0, _update_generalized_link_cost() is useless
-    # link_cost_array may not be correctly set up
-    if not G.has_capi_allocated:
-        G.allocate_for_CAPI()
+    # make sure iteration numbers are both non-negative
+    assert(iter_num>=0)
+    assert(column_update_num>=0)
+
+    # base network
+    A = ui._base_assignment
+
+    links = A.get_links()
+    zones = A.get_zones()
+    ats = A.get_agent_types()
+    dps = A.get_demand_periods()
+    column_pool = A.get_column_pool()
+
+    st = time()
 
     for i in range(iter_num):
         print(f"current iteration number in assignment: {i}")
-        _update_link_travel_time_and_cost(G.link_list)
-        _reset_and_update_link_vol_based_on_columns(G.column_pool, 
-                                                    G.link_list,
-                                                    G.zones,
+        _update_link_travel_time_and_cost(links, ats, dps)
+
+        _reset_and_update_link_vol_based_on_columns(column_pool,
+                                                    links,
+                                                    zones,
+                                                    ats,
+                                                    dps,
                                                     i,
                                                     True)
 
-        for node in G.node_list:
-            _update_generalized_link_cost(G.link_list, G.link_cost_array)
+        # update generalized link cost before assignment
+        _update_generalized_link_cost(A.get_spnetworks())
 
-            single_source_shortest_path(G, node.get_node_id())
-                        
-            _backtrace_shortest_path_tree(node.get_node_no(),
-                                          G.node_list,
-                                          G.link_list,
-                                          G.node_predecessor,
-                                          G.link_predecessor,
-                                          G.node_label_cost,
-                                          G.column_pool,
-                                          i)
+        # loop through all nodes on the base network
+        _assignment(A.get_spnetworks(), column_pool, i)
 
-    _optimize_column_pool(G.column_pool,
-                          G.link_list,
-                          G.zones,
-                          column_update_num)
+    print('\nprocessing time of assignment:{0: .2f}'.format(time()-st)+ 's\n')
 
-    _reset_and_update_link_vol_based_on_columns(G.column_pool, 
-                                                G.link_list,
-                                                G.zones,
+    _optimize_column_pool(column_pool, links, zones, ats, dps, column_update_num)
+
+    _reset_and_update_link_vol_based_on_columns(column_pool,
+                                                links,
+                                                zones,
+                                                ats,
+                                                dps,
                                                 iter_num,
                                                 False)
 
-    _update_link_travel_time_and_cost(G.link_list)
+    _update_link_travel_time_and_cost(links, ats, dps)
 
-    _update_column_travel_time(G.link_list, G.zones, G.column_pool)
\ No newline at end of file
+    _update_column_travel_time(column_pool, links, zones, ats, dps)
\ No newline at end of file
